/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#include "midori.h"

/*
GList* gtk_window_list_toplevels(void)
void gtk_window_present_with_time(GtkWindow*, guint32 timestamp)
*/

// -- Globals variables begin here

GList* browser_list; // Items of type 'CBrowser'
static int browser_count = 0;
CConfig* config;
GtkAccelGroup* accel_group;
GBookmarkFile* bookmarks;
GList* tabtrash; // Items of type 'CResource'

// -- Helper functions begin here

GtkIconTheme* get_icon_theme(GtkWidget* widget)
{
    return gtk_icon_theme_get_for_screen(gtk_widget_get_screen(widget));
}

GtkWidget* menu_item_new(const gchar* sText, const gchar* sIcon
 , const gchar* sPath, const gchar* sKey, GdkModifierType accel_mods
 , GCallback signal, gboolean bSensitive, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* menuitem;
    if(sText)
    {
     menuitem = gtk_image_menu_item_new_with_mnemonic(sText);
     if(sIcon)
     {
      GtkWidget* icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
      GtkIconTheme* icon_theme = get_icon_theme(menuitem);
      if(gtk_icon_theme_has_icon(icon_theme, sIcon))
       gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), icon);
      else
       g_print("Note: The icon %s is not available.", sIcon);
     }
    }
    else
     menuitem = gtk_image_menu_item_new_from_stock(sIcon, NULL);
    if(signal)
     g_signal_connect(menuitem, "activate", GTK_SIGNAL_FUNC(signal), browser);
    gtk_widget_set_sensitive(GTK_WIDGET(menuitem), bSensitive && signal);
    sokoke_menu_item_set_accel(GTK_MENU_ITEM(menuitem), sPath, sKey, accel_mods);
    return menuitem;
}

GtkToolItem* tool_button_new(const gchar* sText, const gchar* sIcon
 , gboolean bImportant, gboolean bSensitive, GCallback signal, gpointer userdata)
{
    DEBUG_FUNC
    GtkToolItem* toolitem;
    if(sText)
    {
     toolitem = gtk_tool_button_new(NULL, sText);
     GtkWidget* icon = gtk_image_new_from_icon_name(sIcon, GTK_ICON_SIZE_MENU);
     GtkIconTheme* icon_theme = get_icon_theme(GTK_WIDGET(toolitem));
     if(gtk_icon_theme_has_icon(icon_theme, sIcon))
      gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(toolitem), sIcon);
     else
      g_print("Note: The icon %s is not available.", sIcon);
    }
    else
     toolitem = gtk_tool_button_new_from_stock(sIcon);
    if(bImportant)
     gtk_tool_item_set_is_important(toolitem, TRUE);
    if(signal)
     g_signal_connect(toolitem, "clicked", GTK_SIGNAL_FUNC(signal), userdata);
    if(!bSensitive || !signal)
     gtk_widget_set_sensitive(GTK_WIDGET(toolitem), FALSE);
    return toolitem;
}

GtkWidget* check_menu_item_new(const gchar* sText
 , const gchar* sPath, const gchar* sKey, GdkModifierType accel_mods
 , GCallback signal, gboolean bSensitive, gboolean bActive, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* menuitem = gtk_check_menu_item_new_with_mnemonic(sText);
    gtk_widget_set_sensitive(menuitem, bSensitive && signal != NULL);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem), bActive);
    if(signal)
     g_signal_connect(menuitem, "activate", G_CALLBACK(signal), browser);
    sokoke_menu_item_set_accel(GTK_MENU_ITEM(menuitem), sPath, sKey, accel_mods);
    return menuitem;
}

GtkWidget* get_current_document(CBrowser* browser)
{
    gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->pages));
    GtkWidget* scrolled = gtk_notebook_get_nth_page(GTK_NOTEBOOK(browser->pages), page);
    return gtk_bin_get_child(GTK_BIN(scrolled));
}

gint get_document_index(GtkWidget* document, CBrowser* browser)
{
    GtkWidget* scrolled = gtk_widget_get_parent(document);
    return gtk_notebook_page_num(GTK_NOTEBOOK(browser->pages), scrolled);
}

void update_favicon(CBrowser* browser)
{
    DEBUG_FUNC
    const gboolean bCurrent = browser->document == get_current_document(browser);
    if(0) //browser->favicon // Has favicon?
    {
     // TODO: use custom icon
     // gtk_image_set_from_file(GTK_IMAGE(browser->icon_page), "image");
     // gtk_image_set_from_file(GTK_IMAGE(browser->page_icon), "image");
    }
    else if(0) // Known mime-type?
    {
     // TODO: Retrieve mime type and load icon; don't forget ftp listings
    }
    else
    {
     gtk_image_set_from_stock(GTK_IMAGE(browser->page_icon), GTK_STOCK_FILE
      , GTK_ICON_SIZE_MENU);
     if(bCurrent)
      gtk_image_set_from_stock(GTK_IMAGE(browser->icon_page), GTK_STOCK_FILE
       , GTK_ICON_SIZE_MENU);
    }
}

void update_security(CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: these are mozilla values, change that!
    //       0 = unknown, 4 = default, 262146 = secure
    // TODO: highlighted entry indicates security, find an alternative
    if(browser->security == 262146)
    {
     gtk_widget_modify_base(browser->toolbar_location, GTK_STATE_NORMAL
      , &browser->toolbar_location->style->base[GTK_STATE_SELECTED]);
     gtk_widget_modify_text(browser->toolbar_location, GTK_STATE_NORMAL
      , &browser->toolbar_location->style->text[GTK_STATE_SELECTED]);
     gtk_image_set_from_stock(GTK_IMAGE(browser->icon_lock)
      , GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_MENU);
    }
    else
    {
     gtk_widget_modify_base(browser->toolbar_location, GTK_STATE_NORMAL
      , NULL);
     gtk_widget_modify_text(browser->toolbar_location, GTK_STATE_NORMAL
      , NULL);
     gtk_image_clear(GTK_IMAGE(browser->icon_lock));
    }
}

void update_visibility(CBrowser* browser, gboolean visibility)
{
    DEBUG_FUNC
    if(gtk_notebook_get_n_pages(GTK_NOTEBOOK(browser->pages)) > 1
     || 1) return;
    // TODO: Check if this is a scripted window, else just return

    // SHOULD SCRIPTS BE ABLE TO HIDE WINDOWS AT ALL?
    if(0 && !visibility)
    {
     gtk_widget_hide(browser->window);
     return;
    }
    else if(!visibility)
     g_print("Window was not hidden.\n");

    if(browser->bMenubar) gtk_widget_show(browser->menubar);
    else gtk_widget_hide(browser->menubar);

    sokoke_widget_set_visible(browser->toolbar, browser->bToolbar);
    sokoke_widget_set_visible(browser->toolbar_location, browser->bLocation);
    sokoke_widget_set_visible(browser->toolbar_search, browser->bLocation);

    if(browser->bStatusbar) gtk_widget_show_all(browser->statusbar);
    else gtk_widget_hide_all(browser->statusbar);

    gtk_window_set_resizable(GTK_WINDOW(browser->window), browser->bResizable);
}

// TODO: Rename this to update_gui_state
void update_toolbar_state(CBrowser* browser)
{
    GtkWidget* document = get_current_document(browser);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_zoom_in)
     , document_can_increase_text_size(document));
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_zoom_out)
     , document_can_decrease_text_size(document));
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_zoom_normal)
     , document_get_text_size(document) != 1);
    gboolean state = document_can_go_back(document);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_go_back), state);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->toolbar_back), state);
    state = document_can_go_forward(document);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_go_forward), state);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->toolbar_forward), state);
}

void update_statusbar_text(CBrowser* browser)
{
    DEBUG_FUNC

    if(browser->status_message)
    {
     gtk_statusbar_pop(GTK_STATUSBAR(browser->statusbar), 1);
     gtk_statusbar_push(GTK_STATUSBAR(browser->statusbar), 1
      , browser->status_message);
    }
    else
    {
     gchar message[256];
     if(browser->loadPercent)
      g_snprintf(message, 255, "d%% loaded, %d/%d bytes"
       , browser->loadPercent, browser->bytesLoaded, browser->maxBytesLoaded);
     else if(browser->bytesLoaded)
      g_snprintf(message, 255, "%d bytes", browser->bytesLoaded);
     else
      g_snprintf(message, 255, " ");
     gtk_progress_bar_set_text(GTK_PROGRESS_BAR(browser->progress), message);
     //g_free(message);
    }
}

void update_feeds(CBrowser* browser)
{
    // TODO: Look for available feeds, requires dom access
}

void update_search_engines(CBrowser* browser)
{
    // TODO: Look for available search engines, requires dom access
}

// FIXME: Rename this to update_status_message
void update_temp_message(CBrowser* browser, const gchar* message)
{
    DEBUG_FUNC
    GString* string = g_string_new(message);
    if(message) g_string_assign(string, message);
    browser->status_message = g_string_free(string, FALSE);
    update_statusbar_text(browser);
}

void update_page_actions(CBrowser* browser)
{
    gboolean bActive = gtk_notebook_get_n_pages(
     GTK_NOTEBOOK(browser->pages)) > 1 ? TRUE : FALSE;
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->pages), bActive);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->menu_closetab), bActive);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->toolbar_closetab), bActive);
    gtk_widget_set_sensitive(GTK_WIDGET(browser->toolbar_tabtrash)
     , g_list_length(tabtrash));
}

gchar* magic_uri(const gchar* sUri)
{
    // Add file:// if we have a local path
    if(g_path_is_absolute(sUri)) return g_strconcat("file://", sUri, NULL);
    // Do we need to add a protocol?
    // TODO: This is not utf8 safe
    if(!strstr(sUri, "://"))
    {
     // TODO: Allow for actual computer names
     // Do we have a domain, ip address or localhost?
     // TODO: This is not utf8 safe
     if(strstr(sUri, ".") != NULL || !strcmp(sUri, "localhost"))
      return g_strconcat("http://", sUri, NULL);
     // Do we have a keyword and a string?
     if(strstr(sUri, " ") != NULL)
      ; // TODO: Search, using smart keywords
     // We only have a word, so search for it
     gchar sSearch[MIDORI_LEN_URI + 1];
     g_snprintf(sSearch, MIDORI_LEN_URI, config->sLocationSearch, sUri);
     return g_strdup(sSearch);
    }
    return g_strdup(sUri);
}

// -- Common (pseudo) signal handlers begin here

void on_document_location_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sUri
#endif
, CBrowser* browser)
{
    DEBUG_FUNC
    if(document == get_current_document(browser))
    {
     #ifndef MIDORI_WEBKIT
     const gchar* sUri = document_get_uri(document);
     #endif
     if(sUri)
     {
      gtk_entry_set_text(GTK_ENTRY(browser->toolbar_location), sUri);
      gtk_label_set_text(GTK_LABEL(browser->page_name), sUri);
     }
     update_temp_message(browser, NULL);
     update_toolbar_state(browser);
    }
}

void on_document_title_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , const gchar* sTitle, const gchar* sUri
#endif
 , CBrowser* browser)
{
    #ifdef MIDORI_WEBKIT
    // TODO: We emulate location_changed here for now
    // Shouldn't we have separated title_changed and location_changed signals?
    on_document_location_changed(document, sUri, browser);
    #endif
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gchar* sTitle = document_get_title(document);
    #endif
    if(sTitle)
    {
     gtk_label_set_text(GTK_LABEL(browser->page_name), sTitle);
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->page_menu))), sTitle);
     if(document == get_current_document(browser))
     {
      gchar* sWindowTitle = g_strconcat(sTitle, " - "
       , g_get_application_name(), NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
      g_free(sWindowTitle);
     }
    }
}

void on_document_load_started(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    if(document == get_current_document(browser))
    {
     gtk_tool_button_set_stock_id(GTK_TOOL_BUTTON(browser->toolbar_refresh)
     , GTK_STOCK_REFRESH);
    }
    update_favicon(browser);
    browser->loadPercent = 0;
    browser->bytesLoaded = 0;
    browser->maxBytesLoaded = 0;
    update_statusbar_text(browser);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress), 0.1);
    gtk_widget_show(browser->progress);
}

void on_document_load_changed(GtkWidget* document
#ifdef MIDORI_WEBKIT
 , gint progress
#else
 , const WebiLoadStatus* status
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    const gint cur = status->received;
    const gint max = status->size;
    #else
    // TODO: The hacks are needed until load_started is availabe
    const gint cur = progress < 100 ? progress : 0;
    const gint max = progress < 100 ? 100 : 0;
    sokoke_widget_set_visible(browser->progress, progress < max);
    #endif
    // Skip null and negative values
    if(max < 1)
    {
     gtk_progress_bar_pulse(GTK_PROGRESS_BAR(browser->progress));
     browser->loadPercent = 0;
     browser->bytesLoaded = cur;
     browser->maxBytesLoaded = 0;
     update_statusbar_text(browser);
    }
    else
    {
     browser->bytesLoaded = cur;
     browser->maxBytesLoaded = max;
     if(cur > max)
      browser->loadPercent = 100;
     else
      browser->loadPercent = (cur * 100) / max;
     gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress)
      , browser->loadPercent / 100.0);
     update_statusbar_text(browser);
    }
}

void on_document_load_finished(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_tool_button_set_stock_id(GTK_TOOL_BUTTON(browser->toolbar_refresh)
     , GTK_STOCK_REFRESH);
    update_favicon(browser);
    browser->loadPercent = 0;
    browser->bytesLoaded = 0;
    browser->maxBytesLoaded = 0;
    update_statusbar_text(browser);
    gtk_widget_hide(browser->progress);
}

void on_document_status_message(GtkWidget* document, gchar* sText
 , CBrowser* browser)
{
    DEBUG_FUNC
    update_temp_message(browser, sText);
}

void on_document_device_changed(GtkWidget* webi
#ifndef MIDORI_WEBKIT
 , WebiDeviceType* device
#endif
 , CBrowser* browser)
{
    // TODO: WHAT is this good for ANYWAY?
    g_print("%s\n", "device_changed");
}

void on_document_selection_changed(GtkWidget* document, CBrowser* browser)
{
    //g_print("%s\n", "selection_changed");
}

void on_document_request_script_prompt(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiPromptArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title;
    GtkMessageType type = GTK_MESSAGE_OTHER;
    GtkButtonsType buttons = GTK_BUTTONS_NONE;
    gint default_response = GTK_RESPONSE_NONE;
    gint response;
    GtkWidget* widget = NULL;

    #ifndef MIDORI_WEBKIT
    switch(args->type)
    {
     case WEBI_ALERT:
      type = GTK_MESSAGE_WARNING;
      buttons = GTK_BUTTONS_CLOSE;
      default_response = GTK_RESPONSE_CLOSE; break;
     case WEBI_CONFIRM:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_YES_NO;
      default_response = GTK_RESPONSE_YES; break;
     case WEBI_INPUT:
      type = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_OK_CANCEL;
      default_response = GTK_RESPONSE_OK;
    }
    #endif

    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, type, buttons
    #ifndef MIDORI_WEBKIT
     , args->msg
    #else
     , ""
    #endif
    ));
    title = g_strconcat("JavaScript - ", document_get_uri(document), NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);

    #ifndef MIDORI_WEBKIT
    if(args->type == WEBI_INPUT)
    #else
    if(0)
    #endif
    {
     widget = gtk_entry_new();
     #ifndef MIDORI_WEBKIT
     gtk_entry_set_text(GTK_ENTRY(widget), args->default_input);
     #endif
     gtk_container_add(GTK_CONTAINER(dialog->vbox), widget);
     gtk_widget_show(widget);
    }

    gtk_dialog_set_default_response(dialog, default_response);
    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifndef MIDORI_WEBKIT
    switch(response)
    {
     case GTK_RESPONSE_YES:
      args->out_ok_pressed = TRUE; break;
     case GTK_RESPONSE_NO:
      args->out_ok_pressed = FALSE; break;
     case GTK_RESPONSE_OK:
      args->out_input = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
    }
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

void on_document_request_auth_prompt(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiAuthArgs* args
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    GtkDialog* dialog;
    gchar* title;
    gchar* msg;
    GtkWidget* hbox;
    GtkWidget* label;
    GtkWidget* user;
    GtkWidget* pass;
    gint response;

    #ifndef MIDORI_WEBKIT
    msg = g_strdup_printf("Authentication required for:\n\n%s", args->realm);
    #endif
    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_OTHER, GTK_BUTTONS_OK_CANCEL
     , msg)); g_free(msg);
    #ifndef MIDORI_WEBKIT
    title = g_strconcat("Authentication - ", args->url, NULL);
    #endif
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);
    #if GTK_CHECK_VERSION(2, 10, 0)
    GtkWidget* image = gtk_image_new_from_icon_name(
     GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
    gtk_widget_show(image);
    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
    #endif
    // TODO: Improve widget alignment.
    hbox = gtk_hbox_new(FALSE, 1);
    label = gtk_label_new("Username");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    user = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), user, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    hbox = gtk_hbox_new(FALSE, 1);
    label = gtk_label_new("Password");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    pass = gtk_entry_new();
    gtk_entry_set_visibility(GTK_ENTRY(pass), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), pass, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_CANCEL);

    gtk_widget_set_sensitive(document, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(document, TRUE);

    #ifndef MIDORI_WEBKIT
    if(response == GTK_RESPONSE_OK)
    {
     args->out_username = g_strdup(gtk_entry_get_text(GTK_ENTRY(user)));
     args->out_password = g_strdup(gtk_entry_get_text(GTK_ENTRY(pass)));
     args->out_ok_pressed = TRUE;
    }
    else
     args->out_ok_pressed = FALSE;
    #endif
    gtk_widget_destroy(GTK_WIDGET(dialog));
}

gboolean on_document_set_cookie(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiCookie* cookie
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    // Whole cookie
    // const gchar* cookie;

    // Cookie name
    // const gchar* name;

    // Cookie value
    // const gchar* value;

    // Cookie comment
    // const gchar* comment;

    // Cookie domain
    // const gchar* domain;

    // Cookie TTL
    // const gchar* maxAge;

    // Cookie path
    // const gchar* path;

    // Is cookie secure?
    // gboolean secure;

    // Cookie-specification version.
    // gint version;

    return TRUE;
}

void on_document_link_hover(GtkWidget* document, const gchar* sTitle
 , const gchar* sLabel, const gchar* sUri, const gchar* sTarget, CBrowser* browser)
{
    DEBUG_FUNC
    update_temp_message(browser, sUri);
    if(sTarget && !g_ascii_strncasecmp(sTarget, "_blank", 6))
     gtk_widget_show(browser->icon_link_window);
    else
     gtk_widget_hide(browser->icon_link_window);
    // TODO: detect "onclick", "onmousedown", "onmouseup" javascript links
    if(sUri && g_str_has_prefix(sUri, "javascript:"))
     gtk_widget_show(browser->icon_link_script);
    else
     gtk_widget_hide(browser->icon_link_script);
}

GtkWidget* on_document_window_open(GtkWidget* document, const gchar* sUri
 , CBrowser* browser)
{
    DEBUG_FUNC
    // A window is created
    // TODO: Find out if this comes from a script or a click
    // TODO: Block scripted popups, return NULL and show status icon
    // Shoud we open a new window or a new tab?
    gboolean bWindow;
    switch(config->iPopupRule)
    {
     case 0: // All popups
         bWindow = TRUE; break;
     case 1: // Only customized popups
         // TODO: Check if window has non-default features
         bWindow = TRUE; break;
     default: // case 2: // No new windows
         bWindow = FALSE;
    }
    CBrowser* newBrowser = browser_new(bWindow ? NULL : browser);
    /*webi_set_group(WEBI(newBrowser->document)
     , webi_get_group(get_current_document(browser)));*/
    return newBrowser->document;
}

void on_document_window_show(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
}

void on_document_window_features(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiWindowProperties* properties
#endif
 , CBrowser* browser)
{
    DEBUG_FUNC
    #ifndef MIDORI_WEBKIT
    browser->bMenubar = TRUE;
    browser->bToolbar = properties->toolbarsVisible;
    browser->bLocation = properties->toolbarsVisible;
    browser->bStatusbar = properties->statusBarVisible;
    //browser->bScrollbars = properties->scrollbarsVisible;
    browser->bResizable = properties->isResizable;
    #endif
    update_visibility(browser, TRUE);
}


void on_document_window_get_size(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiWindowSize* type
#endif
 , GtkAllocation* size, gpointer user_data)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_document_window_set_size(GtkWidget* document
#ifndef MIDORI_WEBKIT
 , WebiWindowSize* type
#endif
 , GtkRequisition* size, CBrowser* browser)
{
    DEBUG_FUNC
    UNIMPLEMENTED
}

void on_document_window_close(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: Close window if opened from script
    UNIMPLEMENTED
}

// Document actions begin here

#define DOCUMENT_IS_SENSITIVE(rv) \
 if(!GTK_WIDGET_SENSITIVE(document)) return rv

void document_load_uri(GtkWidget* document, const gchar* sUri)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    // TODO: Check if we want an external app for this uri scheme
    #ifdef MIDORI_WEBI
    webi_load_url(WEBI(document), sUri);
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_open(WEBKIT_GTK_PAGE(document), (gchar*)sUri);
    #else
    WebkitFrame* frame = webkit_view_get_main_frame(WEBKIT_VIEW(document));
    webkit_frame_load_http_get_str(frame, sUri);
    #endif
}

void document_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_backward(WEBKIT_GTK_PAGE(document));
    #else
    webkit_view_go_back(WEBKIT_VIEW(document));
    #endif
}

void document_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_go_forward(WEBKIT_GTK_PAGE(document));
    #else
    webkit_view_go_forward(WEBKIT_VIEW(document));
    #endif
}

void document_refresh(GtkWidget* document, gboolean bSkipCache)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_refresh(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_reload(WEBKIT_GTK_PAGE(document));
    #else
    WebkitFrame* frame = webkit_view_get_main_frame(WEBKIT_VIEW(document));
    webkit_frame_reload(frame);
    #endif
}

void document_stop(GtkWidget* document)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_stop_load(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    webkit_gtk_page_stop_loading(WEBKIT_GTK_PAGE(document));
    #else
    WebkitFrame* frame = webkit_view_get_main_frame(WEBKIT_VIEW(document));
    webkit_frame_stop_loading(frame);
    #endif
}

gboolean document_can_go_back(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return FALSE;
    #ifdef MIDORI_WEBI
    return webi_can_go_back(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_backward(WEBKIT_GTK_PAGE(document));
    #else
    return webkit_view_can_go_back(WEBKIT_VIEW(document));
    #endif
}

gboolean document_can_go_forward(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return FALSE;
    #ifdef MIDORI_WEBI
    return webi_can_go_forward(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    return webkit_gtk_page_can_go_forward(WEBKIT_GTK_PAGE(document));
    #else
    return webkit_view_can_go_forward(WEBKIT_VIEW(document));
    #endif
}

G_CONST_RETURN gchar* document_get_uri(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return NULL;
    #ifdef MIDORI_WEBI
    return webi_get_location(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return ""; //webkit_gtk_page_get_uri(WEBKIT_GTK_PAGE(document));
    #else
    return webkit_view_get_main_frame_url(WEBKIT_VIEW(document));
    #endif
}

G_CONST_RETURN gchar* document_get_title(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return NULL;
    #ifdef MIDORI_WEBI
    return webi_get_title(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    // TODO: This hack is needed until webkit api has it
    const gchar* sTitle;
    if(GTK_IS_WIDGET(document))
    {
     GtkWidget* scrolled = gtk_widget_get_parent(document);
     GtkWidget* notebook = gtk_widget_get_parent(scrolled);
     GtkWidget* ebox = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), scrolled);
     GList* children = gtk_container_get_children(GTK_CONTAINER(ebox));
     GtkWidget* hbox = (GtkWidget*)g_list_nth_data(children, 0);
     g_list_free(children);
     children = gtk_container_get_children(GTK_CONTAINER(hbox));
     GtkWidget* label = (GtkWidget*)g_list_nth_data(children, 1);
     g_list_free(children);
     sTitle = gtk_label_get_text(GTK_LABEL(label));
    }
    /*UNIMPLEMENTED*/ return sTitle; //webkit_gtk_page_get_title(WEBKIT_GTK_PAGE(document));
    #else
    return webkit_view_get_main_frame_title(WEBKIT_VIEW(document));
    #endif
}

// TODO: This better be document_find(document, text, forward, case, wrap)
gboolean document_find(GtkWidget* document, const gchar* sText
 , gboolean bCase, gboolean bForward)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBI
    return webi_find(WEBI(document), sText, bCase, bForward);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; /*webkit_gtk_page_search_for(WEBKIT_GTK_PAGE(document)
     , sText, bForward, bCase, TRUE);*/
    #else
    return webkit_view_search_for(WEBKIT_VIEW(document), sText
     , bCase, bForward, TRUE);
    #endif
}

gboolean document_can_increase_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return FALSE;
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_increase_text_size(WEBKIT_GTK_PAGE(document));
    #else
    return TRUE;
    #endif
}

gboolean document_can_decrease_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return FALSE;
    #ifdef MIDORI_WEBI
    return TRUE;
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return FALSE; //webkit_gtk_page_can_decrease_text_size(WEBKIT_GTK_PAGE(document));
    #else
    return TRUE;
    #endif
}

void document_set_text_size(GtkWidget* document, gfloat iSize)
{
    DEBUG_FUNC
    DOCUMENT_IS_SENSITIVE();
    #ifdef MIDORI_WEBI
    webi_set_text_multiplier(WEBI(document), iSize);
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED //webkit_gtk_page_set_text_size_multiplier(WEBKIT_GTK_PAGE(document), iSize);
    #else
    webkit_view_set_text_multiplier(WEBKIT_VIEW(document), iSize);
    #endif
}

gfloat document_get_text_size(GtkWidget* document)
{
    DEBUG_FUNC
    if(!document) return 1;
    #ifdef MIDORI_WEBI
    return webi_get_text_multiplier(WEBI(document));
    #elif defined(MIDORI_WEBKIT)
    UNIMPLEMENTED return 1;//webkit_gtk_page_get_text_size_multiplier(WEBKIT_GTK_PAGE(document));
    #else
    return webkit_view_get_text_multiplier(WEBKIT_VIEW(document));
    #endif
}

void document_close(GtkWidget* document, CBrowser* browser)
{
    DEBUG_FUNC
    CResource* resource = g_new0(CResource, 1);
    resource->sUri = g_strdup(document_get_uri(document));
    resource->sTitle = g_strdup(document_get_title(document));
    if(!resource->sUri || !resource->sTitle)
    {
     // For now we add everything to the list
     // Later we should skip e.g. blank pages
     resource->sUri = g_strdup("about:blank");
     resource->sTitle = g_strdup("Blank");
    }
    tabtrash = g_list_prepend(tabtrash, resource);
    // Remove last item if we already have 10
    if(g_list_length(tabtrash) > 10)
     tabtrash = g_list_remove(tabtrash, g_list_last(tabtrash)->data);
    gtk_widget_destroy(browser->page_menu);
    gtk_notebook_remove_page(GTK_NOTEBOOK(browser->pages)
     , gtk_notebook_page_num(GTK_NOTEBOOK(browser->pages)
     , document));
    update_page_actions(browser);
}

#undef DOCUMENT_IS_SENSITIVE

// -- GTK+ signal handlers begin here

void on_menu_file_new_window_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    browser_new(NULL);
}

void on_menu_file_new_tab_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    browser_new(browser);
    update_page_actions(browser);
}

void on_menu_file_close_tab_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    document_close(get_current_document(browser), browser);
}

void on_menu_file_close_window_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gtk_widget_destroy(GTK_WIDGET(browser->window));
}

void on_menu_file_quit_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gtk_main_quit();
}

void on_menu_edit_find_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    if(GTK_WIDGET_VISIBLE(browser->findbox))
     gtk_widget_hide(browser->findbox);
    else
    {
     gtk_widget_show(browser->findbox);
     gtk_widget_grab_focus(GTK_WIDGET(browser->findbox_text));
    }
}

void on_menu_view_show_toolbar_activate(GtkCheckMenuItem* widget, CBrowser* browser)
{
    config->bShowToolbar = gtk_check_menu_item_get_active(widget);
    sokoke_widget_set_visible(browser->toolbar, config->bShowToolbar);
}

void on_menu_view_show_bookmarkbar_activate(GtkCheckMenuItem* widget, CBrowser* browser)
{
    config->bShowBookmarkbar = gtk_check_menu_item_get_active(widget);
    sokoke_widget_set_visible(browser->bookmarkbar, config->bShowBookmarkbar);
}

void on_menu_view_show_downloadbar_activate(GtkCheckMenuItem* widget, CBrowser* browser)
{
    /*config->bShowDownloadbar = gtk_check_menu_item_get_active(widget);
    sokoke_widget_set_visible(browser->downloadbar, config->bShowDownloadbar);*/
}

void on_menu_view_show_statusbar_activate(GtkCheckMenuItem* widget, CBrowser* browser)
{
    config->bShowStatusbar = gtk_check_menu_item_get_active(widget);
    sokoke_widget_set_visible(browser->statusbar, config->bShowStatusbar);
}

void on_menu_view_refresh_activate(GtkWidget* widget, CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
    document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
}

void on_menu_view_stop_activate(GtkWidget* widget, CBrowser* browser)
{
    document_stop(get_current_document(browser));
}

void on_menu_view_zoom_in_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom + 0.1);
}

void on_menu_view_zoom_out_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    const gfloat zoom = document_get_text_size(get_current_document(browser));
    document_set_text_size(get_current_document(browser), zoom - 0.1);
}

void on_menu_view_zoom_default_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    document_set_text_size(get_current_document(browser), 1);
}

void on_menu_go_back_activate(GtkWidget* widget, CBrowser* browser)
{
    document_go_back(get_current_document(browser));
}

void on_menu_go_forward_activate(GtkWidget* widget, CBrowser* browser)
{
    document_go_forward(get_current_document(browser));
}

void on_menu_go_home_activate(GtkWidget* widget, CBrowser* browser)
{
    document_load_uri(get_current_document(browser), config->sHomepage);
}

void on_menu_tools_colorpicker_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    // TODO: Prevent the user from opening multiple color pickers.
    GtkWidget* color_picker = gtk_color_selection_dialog_new("Color picker");
    gtk_window_set_type_hint(GTK_WINDOW(color_picker), GDK_WINDOW_TYPE_HINT_UTILITY);
    gtk_window_set_icon_name(GTK_WINDOW(color_picker), GTK_STOCK_COLOR_PICKER);
    gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(color_picker)->ok_button);
    gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(color_picker)->cancel_button);
    g_signal_connect(GTK_OBJECT(color_picker), "response"
     , GTK_SIGNAL_FUNC(gtk_widget_destroy), color_picker);
    gtk_widget_show(color_picker);
}

void on_menu_window_item_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    gint page = get_document_index(browser->document, browser);
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->pages), page);
}

void on_menu_help_about_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    gtk_show_about_dialog(GTK_WINDOW(browser->window)
     //, "title", _("About %s")
     , "logo-icon-name", "web-browser"
     , "name", PACKAGE_NAME
     , "version", PACKAGE_VERSION
     , "comments", "A lightweight web browser."
     , "copyright", "Copyright © 2007 Christian Dywan"
     , "website", "http://software.twotoasts.de"
     , "authors", credits_authors
     , "documenters", credits_documenters
     , "artists", credits_artists
     , "license", license_gpl_2
     , "wrap-license", TRUE
     //, "translator-credits", _("translator-credits")
     , NULL);
}

void on_toolbar_refresh_activate(GtkWidget* widget, CBrowser* browser)
{
    // Reload or stop loading, whatever is needed
    const gchar* sIcon = gtk_tool_button_get_stock_id(GTK_TOOL_BUTTON(
     browser->toolbar_refresh));
    if(!strcmp(sIcon, GTK_STOCK_REFRESH))
    {
     GdkModifierType state = (GdkModifierType)0;
     gint x, y; gdk_window_get_pointer(NULL, &x, &y, &state);
     document_refresh(get_current_document(browser), state & GDK_SHIFT_MASK);
    }
    else
     document_stop(get_current_document(browser));
}

gboolean on_toolbar_entry_location_key_down(GtkWidget* widget
 , GdkEventKey* event, CBrowser* browser)
{
    DEBUG_FUNC
    const gchar* sUri;
    switch(event->keyval)
    {
     case GDK_Return:
      sUri = gtk_entry_get_text(GTK_ENTRY(widget));
      if(sUri != NULL)
      {
       gchar* sNewUri = magic_uri(sUri);
       // TODO: add to auto completion
       // gtk_list_store_insert (completion_store, &iter, 0);
       // gtk_list_store_set (completion_store, &iter, 0, url, -1);
       document_load_uri(get_current_document(browser), sNewUri);
       g_free(sNewUri);
      }
      return TRUE;
     case GDK_Escape:
      sUri = document_get_uri(get_current_document(browser));
      gtk_entry_set_text(GTK_ENTRY(widget), sUri);
      return TRUE;
    }
    return FALSE;
}

void on_toolbar_entry_search_activate(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: use custom search engine
    // TODO: add to auto completion
    gchar* sText = g_strconcat("http://www.google.com/search?q="
     , gtk_entry_get_text(GTK_ENTRY(widget)), NULL);
    document_load_uri(get_current_document(browser), sText);
    g_free(sText);
}

void on_menu_go_tabtrash_clear_activate(GtkImageMenuItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: What is wrong here?
    g_list_foreach(tabtrash, (GFunc)g_free, NULL); g_list_free(tabtrash);
    update_page_actions(browser);
}

void on_toolbar_button_tabtrash_clicked(GtkToolItem* widget, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* menu = gtk_menu_new();
    GList* items = g_list_first(tabtrash);
    guint i; CResource* resource;
    for(i = 0; i < g_list_length(items); i++)
    {
     resource = (CResource*)g_list_nth_data(items, i);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(resource->sTitle
      , GTK_STOCK_FILE, "", NULL, GDK_CONTROL_MASK
      , NULL/*on_menu_go_tabtrash_item_activate*/, TRUE, browser));
    }
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Clear closed tabs list"
     , GTK_STOCK_CLEAR, "", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_go_tabtrash_clear_activate), TRUE, browser));
    g_signal_connect(menu, "deactivate", G_CALLBACK(gtk_widget_destroy), NULL);
    gtk_widget_show_all(menu);

    gtk_menu_attach_to_widget(GTK_MENU(menu), GTK_WIDGET(widget), NULL);
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 0
     , gtk_get_current_event_time());
}

void on_sidebars_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    gtk_widget_hide(browser->sidebars);
}

gint on_notebook_tab_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    if(event->button == 1 && event->type == GDK_2BUTTON_PRESS)
    {
     // Toggle the label visibility on double click
     GtkWidget* child = gtk_bin_get_child(GTK_BIN(widget));
     GList* children = gtk_container_get_children(GTK_CONTAINER(child));
     child = (GtkWidget*)g_list_nth_data(children, 1);
     gboolean visible = gtk_widget_get_child_visible(GTK_WIDGET(child));
     gtk_widget_set_child_visible(GTK_WIDGET(child), !visible);
     gint a, b; sokoke_widget_get_text_size(browser->page_name, "M", &a, &b);
     gtk_widget_set_size_request(child, !visible
      ? a * config->iTabSize : 0, !visible ? -1 : 0);
     g_list_free(children);
     return TRUE;
    }
    else if(event->button == 2)
    {
     // Close the document on middle click
     // TODO: This won't close the right document, will it?
     document_close(browser->document, browser);
     return TRUE;
    }

    return FALSE;
}

gint on_notebook_tab_close_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: This won't close the right document, will it?
    document_close(browser->document, browser);

    return FALSE;
}

void on_notebook_switch_page(GtkWidget* widget, GtkNotebookPage* page
 , guint page_num, CBrowser* browser)
{
    DEBUG_FUNC
    GtkWidget* docbox = gtk_notebook_get_nth_page(GTK_NOTEBOOK(browser->pages)
     , page_num);
    GtkWidget* document = gtk_bin_get_child(GTK_BIN(docbox));
    const gchar* sUri = document_get_uri(document);
    if(sUri != NULL) gtk_entry_set_text(GTK_ENTRY(browser->toolbar_location), sUri);
    const gchar* sTitle = document_get_title(document);
    if(sTitle != NULL)
    {
     gchar* sWindowTitle = g_strconcat(sTitle, " - "
      , g_get_application_name(), NULL);
     gtk_window_set_title(GTK_WINDOW(browser->window), sWindowTitle);
     g_free(sWindowTitle);
    }
    update_favicon(browser);
    update_security(browser);
    update_toolbar_state(browser);
    update_statusbar_text(browser);
    update_feeds(browser);
    update_search_engines(browser);
}

gboolean on_document_mouse_down(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    switch(event->button)
    {
     case 1:
      if(state & GDK_SHIFT_MASK)
      {
       // TODO: Open link in new window
       UNIMPLEMENTED return TRUE;
      }
      else if(state & GDK_MOD1_MASK)
      {
       // TODO: Open link in new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 2:
      if(state & GDK_CONTROL_MASK)
      {
       document_set_text_size(get_current_document(browser), 1);
       return TRUE;
      }
      else
      {
       // TODO: Middle click opens new tab
       UNIMPLEMENTED return TRUE;
      }
      break;
     case 3:
      on_document_popup(get_current_document(browser), event, browser);
      return TRUE;
    }
    return FALSE;
}

void on_document_popup(GtkWidget* document, GdkEventButton* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    // TODO: There seems to be an issue with webkitgtk - if you see it at all :/
    GtkWidget* menu = gtk_menu_new();
    gboolean bLink = FALSE; // TODO: Did we right-click a link?
    gboolean bImage = FALSE; // TODO: Did we right-click an image?
    gchar* sSelection = NULL; // TODO: The possibly selected text
    gchar* sString; // this is for g_strdup_printf
    if(!bLink && !bImage && !sSelection)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
      , GTK_STOCK_GO_BACK, "Go/Back", NULL, GDK_CONTROL_MASK
      , G_CALLBACK(on_menu_go_back_activate), document_can_go_back(document), browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
      , GTK_STOCK_GO_FORWARD, "Go/Forward", NULL, GDK_CONTROL_MASK
      , G_CALLBACK(on_menu_go_forward_activate), document_can_go_forward(document), browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_REFRESH, "View/Refresh", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_refresh_activate), TRUE, browser));
     // TODO: Is appointment-new a good choice?
     // TODO: What if it isn't available?
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Refresh every >"
     , "appointment-new", "View/Refresh every", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_refresh_every_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_STOP, "Go/Stop", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_stop_activate), TRUE, browser));
    }
    if(bLink)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Open link in new tab"
     , STOCK_TAB_NEW, "Go/Open link in new tab", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Open link in current tab"
     , NULL, "Go/Open link in current tab", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Open link in new window"
     , STOCK_WINDOW_NEW, "Go/Open link in new window", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
    }
    if(sSelection != NULL)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
     if(1) // TODO: Are there any links at all?
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Open links in tabs"
      , NULL, "Go/Undo close tab", NULL, GDK_CONTROL_MASK
      , NULL/*on_menu_view_source_activate*/, FALSE, browser));
     sString = g_strdup_printf("Open \"%s\" in new tab", sSelection);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(sString
     , STOCK_TAB_NEW, "Go/Open string in new tab", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser)); g_free(sString);
     sString = g_strdup_printf("Open \"%s\" in current tab", sSelection);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(sString
     , NULL, "Go/Open string in current tab", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser)); g_free(sString);
     sString = g_strdup_printf("Open \"%s\" in new window", sSelection);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(sString
     , STOCK_WINDOW_NEW, "Go/Open string in new window", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser)); g_free(sString);
    }
    if(!bLink && !bImage && !sSelection)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Undo close tab"
     , GTK_STOCK_UNDELETE, "Go/Undo close tab", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Bookmark this page..."
     , STOCK_BOOKMARK_NEW, "Bookmarks/Bookmark", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_bookmarks_bookmark_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_SAVE_AS, "File/Save as", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_file_save_as_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PRINT, "File/Print", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_file_print_activate*/, TRUE, browser));
    }
    if(sSelection != NULL)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_COPY, "Edit/Copy", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_edit_copy_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_SELECT_ALL, "View/Select all", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_select_all_activate*/, TRUE, browser));
     // TODO: Use selected search engine's icon
     sString = g_strdup_printf("Search for \"%s\"", sSelection);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(sString
     , GTK_STOCK_FIND, "Edit/Search for", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_go_search_for_activate*/, TRUE, browser)); g_free(sString);
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Search with >"
     , GTK_STOCK_FIND, "Edit/Search with", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Selection source"
     , STOCK_SOURCE_VIEW, "View/Selection source", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_selection_source_activate*/, TRUE, browser));
    }
    if(!bLink && !bImage && !sSelection)
    {
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Background image"
     , STOCK_IMAGE, "View/Background image", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_background_image_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_SOURCE_VIEW, "View/Source", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
     gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PROPERTIES, "View/Properties", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_properties_activate*/, TRUE, browser));
    }
    g_signal_connect(menu, "deactivate", G_CALLBACK(gtk_widget_destroy), NULL);
    gtk_widget_show_all(menu);

    int button, event_time;
    if(event)
    {
     button = event->button;
     event_time = event->time;
    }
    else
    {
     button = 0;
     event_time = gtk_get_current_event_time();
    }

    gtk_menu_attach_to_widget(GTK_MENU(menu), document, NULL);
    gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, button, event_time);
}

gboolean on_document_scroll(GtkWidget* document, GdkEventScroll* event
 , CBrowser* browser)
{
    DEBUG_FUNC
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    if(state & GDK_CONTROL_MASK)
    {
     const gfloat size = document_get_text_size(get_current_document(browser));
     if(event->direction == GDK_SCROLL_DOWN)
      document_set_text_size(get_current_document(browser), size + 0.1);
     else if(event->direction == GDK_SCROLL_UP)
      document_set_text_size(get_current_document(browser), size - 0.1);
     return TRUE;
    }
    else
     return FALSE;
}

void on_document_destroy(GtkWidget* widget, CBrowser* browser)
{
    DEBUG_FUNC
    // Update browser list, free memory and possibly quit
    browser_count--;
    GList* tmp_list = g_list_find(browser_list, browser);
    browser_list = g_list_remove_link(browser_list, tmp_list);
    g_free(browser->status_message);
    g_strfreev(browser->resources);
    if(!browser_count)
    {
     // TODO: What is wrong here?
     g_list_foreach(tabtrash, (GFunc)g_free, NULL); g_list_free(tabtrash);
     gtk_main_quit();
    }
}

void on_findbox_button_close_clicked(GtkWidget* widget, CBrowser* browser)
{
    gtk_widget_hide(browser->findbox);
}

void on_findbox_button_down_clicked(GtkWidget* widget, CBrowser* browser)
{
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, TRUE);
}

void on_findbox_button_up_clicked(GtkWidget* widget, CBrowser* browser)
{
    const gchar* sText = gtk_entry_get_text(GTK_ENTRY(browser->findbox_text));
    const gboolean bCase = gtk_toggle_tool_button_get_active(
     GTK_TOGGLE_TOOL_BUTTON(browser->findbox_case));
    document_find(get_current_document(browser), sText, bCase, FALSE);
}

gboolean on_window_destroy(GtkWidget* widget, GdkEvent* event, CBrowser* browser)
{
    DEBUG_FUNC
    gboolean bProceed = TRUE;
    // TODO: What if there are multiple windows?
    // TODO: Smart dialog, à la 'Session?: Save, Discard, Cancel'
    // TODO: Pref startup: session, ask, homepage, blank <-- ask
    // TODO: Pref quit: session, ask, none <-- ask

    if(0 /*browser_count > 1*/)
    {
     GtkDialog* dialog;
     dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
      , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO
      , "There is more than one tab open. Do you want to close anyway?"));
     gtk_window_set_title(GTK_WINDOW(dialog), g_get_application_name());

     gtk_dialog_set_default_response(dialog, GTK_RESPONSE_YES);
     bProceed = gtk_dialog_run(dialog) == GTK_RESPONSE_YES;

     gtk_widget_destroy(GTK_WIDGET(dialog));
    }
    return !bProceed;
}

// -- Browser creation begins here

static CBrowser* browser_new(CBrowser* oldBrowser)
{
    DEBUG_FUNC
    browser_count++;
    CBrowser* browser = 0;
    browser = g_new0(CBrowser, 1);
    browser_list = g_list_prepend(browser_list, browser);

    if(!oldBrowser) {

    GtkWidget* label; GtkWidget* hbox;

    // Setup the window
    browser->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    //sokoke_widget_set_transparency(browser->window, 0.7);
    GdkScreen* screen = gtk_window_get_screen(GTK_WINDOW(browser->window));
    gdouble width = gdk_screen_get_width(screen) / 1.7;
    gdouble height = gdk_screen_get_height(screen) / 1.7;
    gtk_window_set_default_size(GTK_WINDOW(browser->window), (gint)width, (gint)height);
    gtk_window_set_title(GTK_WINDOW(browser->window), g_get_application_name());
    gtk_window_add_accel_group(GTK_WINDOW(browser->window), accel_group);
    g_signal_connect(GTK_OBJECT(browser->window), "delete-event"
     , G_CALLBACK(on_window_destroy), browser);
    GtkWidget* vbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(browser->window), vbox);
    gtk_widget_show(vbox);

    // Menu "File"
    browser->menubar = gtk_menu_bar_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->menubar, FALSE, FALSE, 0);
    GtkWidget* menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    GtkWidget* menuitem = gtk_menu_item_new_with_mnemonic("_File");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_WINDOW_NEW, "File/New window", "n", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_file_new_window_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_TAB_NEW, "File/New tab", "t", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_file_new_tab_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL, GTK_STOCK_OPEN
     , "File/Open", "o", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_file_open_file_activate*/, TRUE, browser));
    browser->menu_closetab = menu_item_new(NULL
     , STOCK_TAB_CLOSE, "File/Close tab", "w", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_file_close_tab_activate), FALSE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_closetab);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_WINDOW_CLOSE, "File/Close window", "F4", GDK_MOD1_MASK
     , G_CALLBACK(on_menu_file_close_window_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Page setup"
     , GTK_STOCK_PROPERTIES, "File/Page setup", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_file_print_setup_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PRINT_PREVIEW, "File/Print preview", NULL, (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_file_print_preview_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PRINT, "File/Print", "p", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_file_print_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_QUIT, "File/Quit", "q", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_file_quit_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Edit"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_EDIT, NULL);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), NULL);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_UNDO, "Edit/Undo", "z", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_undo_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_REDO, "Edit/Redo", "y", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_redo_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_CUT, "Edit/Cut", "x", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_cut_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_COPY, "Edit/Copy", "c", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_copy_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PASTE, "Edit/Paste", "v", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_paste_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_DELETE, "Edit/Delete", "Delete", (GdkModifierType)0
     , NULL/*on_menu_edit_delete_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_SELECT_ALL, "Edit/Select all", "a", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_select_all_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_FORM_FILL, "Edit/Formfill", NULL, (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_edit_formfill_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_FIND, "Edit/Find", "f", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_edit_find_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PREFERENCES, "Edit/Settings", NULL, (GdkModifierType)(GDK_CONTROL_MASK | GDK_SHIFT_MASK)
     , NULL/*on_menu_edit_settings_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "View"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_menu_item_new_with_mnemonic("_View");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), check_menu_item_new("Show menubar"
     , "View/Show menubar", NULL, (GdkModifierType)0
     , NULL/*G_CALLBACK(on_menu_view_show_menubar_activate)*/, TRUE, TRUE, browser));
      // Submenu "Toolbars"
      GtkWidget* submenu = gtk_menu_new();
      GtkWidget* submenuitem = gtk_menu_item_new_with_mnemonic("_Toolbars");
      gtk_menu_item_set_submenu(GTK_MENU_ITEM(submenuitem), submenu);
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("Navigation"
       , "View/Toolbar", NULL, (GdkModifierType)0
       , G_CALLBACK(on_menu_view_show_toolbar_activate)
       , TRUE, config->bShowToolbar, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("Bookmarkbar"
       , "View/Bookmarkbar", NULL, (GdkModifierType)0
       , G_CALLBACK(on_menu_view_show_bookmarkbar_activate)
       , TRUE, config->bShowBookmarkbar, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("Downloadbar"
       , "View/Downloadbar", NULL, (GdkModifierType)0
       , G_CALLBACK(on_menu_view_show_downloadbar_activate)
       , FALSE/*TRUE*/, FALSE, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("Statusbar"
       , "View/Statusbar", NULL, (GdkModifierType)0
       , G_CALLBACK(on_menu_view_show_statusbar_activate)
       , TRUE, config->bShowStatusbar, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), submenuitem);
      // Submenu "Sidebars"
      submenu = gtk_menu_new();
      submenuitem = gtk_menu_item_new_with_mnemonic("_Sidebars");
      gtk_menu_item_set_submenu(GTK_MENU_ITEM(submenuitem), submenu);
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("Bookmarks"
       , "View/Bookmarks", NULL, (GdkModifierType)0
       , NULL/*on_menu_view_show_bookmarks_activate*/, TRUE, FALSE, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(submenu), check_menu_item_new("History"
       , "View/History", NULL, (GdkModifierType)0
       , NULL/*on_menu_view_show_history_activate*/, TRUE, FALSE, browser));
      gtk_menu_shell_append(GTK_MENU_SHELL(menu), submenuitem);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_REFRESH, "View/Refresh", "r", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_refresh_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_STOP, "View/Stop", "Escape", (GdkModifierType)0
     , G_CALLBACK(on_menu_view_stop_activate), TRUE, browser));
    browser->menu_zoom_in = menu_item_new(NULL, GTK_STOCK_ZOOM_IN
     , "View/Zoom in", "plus", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_zoom_in_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_zoom_in);
    browser->menu_zoom_out = menu_item_new(NULL, GTK_STOCK_ZOOM_OUT
     , "View/Zoom out", "minus", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_zoom_out_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_zoom_out);
    browser->menu_zoom_normal = menu_item_new(NULL, GTK_STOCK_ZOOM_100
     , "View/Zoom normal", "0", (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_view_zoom_default_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_zoom_normal);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_SOURCE_VIEW, "View/Source", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_source_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_PROPERTIES, "View/Properties", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_view_properties_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Go"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_menu_item_new_with_mnemonic("_Go");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    browser->menu_go_back = menu_item_new(NULL, GTK_STOCK_GO_BACK
     , "Go/Back", "Left", (GdkModifierType)GDK_MOD1_MASK
     , G_CALLBACK(on_menu_go_back_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_go_back);
    browser->menu_go_forward = menu_item_new(NULL, GTK_STOCK_GO_FORWARD
     , "Go/Forward", "Right", (GdkModifierType)GDK_MOD1_MASK
     , G_CALLBACK(on_menu_go_forward_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), browser->menu_go_forward);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_HOMEPAGE, "Go/Home", "Home", (GdkModifierType)GDK_MOD1_MASK
     , G_CALLBACK(on_menu_go_home_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Location..."
     , GTK_STOCK_JUMP_TO, "Go/Location", "l", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_go_location_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    // TODO: Insert closed tabs here
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Closed tabs"
     , STOCK_USER_TRASH, "Go/Closed tabs", NULL, (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_go_tabtrash_activate*/, FALSE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Bookmarks"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_menu_item_new_with_mnemonic("_Bookmarks");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_BOOKMARK_NEW, "Bookmarks/Add bookmark", "d", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_bookmarks_add_bookmark_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_BOOKMARKS, "Bookmarks/Edit bookmarks", "b", (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_bookmarks_edit_bookmarks_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Folder"
     , GTK_STOCK_DIRECTORY, "", NULL, (GdkModifierType)0
     , NULL/*on_menu_bookmarks_item_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Example"
     , STOCK_BOOKMARK, "", NULL, (GdkModifierType)0
     , NULL/*on_menu_bookmarks_item_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Example"
     , STOCK_BOOKMARK, "", NULL, (GdkModifierType)0
     , NULL/*on_menu_bookmarks_item_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Tools"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_menu_item_new_with_mnemonic("_Tools");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_DOWNLOADS, "Tools/Downloads", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_tools_downloads_activate*/, TRUE, browser));
    // TODO: STOCK_THEME feels wrong here
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Userstyles"
     , STOCK_THEME, "Tools/Userstyles", NULL, (GdkModifierType)GDK_CONTROL_MASK
     , NULL/*on_menu_tools_userscripts_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Userscripts"
     , STOCK_SCRIPT, "Tools/Userscripts", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_tools_userscripts_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_EXTENSIONS, "Tools/Extensions", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_tools_extensions_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_PLUGINS, "Tools/Plugins", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_tools_plugins_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Color picker"
     , GTK_STOCK_COLOR_PICKER, "Tools/Color picker", NULL, (GdkModifierType)GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_tools_colorpicker_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , STOCK_ERROR_CONSOLE, "Tools/Error console", NULL, (GdkModifierType)(GDK_CONTROL_MASK
     | GDK_SHIFT_MASK), NULL/*on_menu_tools_error_console_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    // TODO: This is reserved for widgets and extensions
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Window"
    browser->menu_window = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(browser->menu_window), accel_group);
    menuitem = gtk_menu_item_new_with_mnemonic("_Window");
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), menu_item_new("Load session"
     , NULL, "Window/Load session", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_file_load_session_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), menu_item_new("Save session"
     , NULL, "Window/Save session", NULL, GDK_CONTROL_MASK
     , NULL/*on_menu_file_save_session_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), gtk_separator_menu_item_new());
    // All open documents shall be represented here
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), browser->menu_window);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Menu "Help"
    menu = gtk_menu_new();
    gtk_menu_set_accel_group(GTK_MENU(menu), accel_group);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_HELP, NULL);
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(menuitem), NULL);
    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new("Contents"
     , GTK_STOCK_HELP, "Help/Contents", "F1", (GdkModifierType)0
     , NULL/*on_menu_help_contents_activate*/, TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), gtk_separator_menu_item_new());
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), menu_item_new(NULL
     , GTK_STOCK_ABOUT, "Help/About", NULL, GDK_CONTROL_MASK
     , G_CALLBACK(on_menu_help_about_activate), TRUE, browser));
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menubar), menuitem);

    // Toolbar
    browser->toolbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->toolbar), (GtkIconSize)config->iToolbarSize);
    gtk_toolbar_set_show_arrow(GTK_TOOLBAR(browser->toolbar), TRUE);

    GtkToolItem* toolitem = tool_button_new(NULL
     , STOCK_TAB_NEW, TRUE, TRUE, G_CALLBACK(on_menu_file_new_tab_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), toolitem, -1);
    browser->toolbar_back = tool_button_new(NULL
     , GTK_STOCK_GO_BACK, FALSE, FALSE, G_CALLBACK(on_menu_go_back_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), browser->toolbar_back, -1);
    browser->toolbar_forward = tool_button_new(NULL
     , GTK_STOCK_GO_FORWARD, FALSE, FALSE, G_CALLBACK(on_menu_go_forward_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), browser->toolbar_forward, -1);
    browser->toolbar_refresh = tool_button_new(NULL
     , GTK_STOCK_REFRESH, FALSE, TRUE, G_CALLBACK(on_toolbar_refresh_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), browser->toolbar_refresh, -1);
    toolitem = tool_button_new(NULL, STOCK_HOMEPAGE
     , FALSE, TRUE, G_CALLBACK(on_menu_go_home_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), toolitem, -1);
    toolitem = tool_button_new(NULL, STOCK_FORM_FILL
     , TRUE, TRUE, NULL/*on_menu_go_formfill_activate*/, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), toolitem, -1);

    // Location entry
    browser->toolbar_location = sexy_icon_entry_new();
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->toolbar_location), TRUE);
    browser->icon_page = gtk_image_new_from_stock(GTK_STOCK_FILE, GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->toolbar_location)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_page));
    browser->icon_lock = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->toolbar_location)
     , SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE(browser->icon_lock));
    g_signal_connect(browser->toolbar_location, "key-press-event"
     , G_CALLBACK(on_toolbar_entry_location_key_down), browser);
    toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(GTK_TOOL_ITEM(toolitem), TRUE);
    gtk_container_add(GTK_CONTAINER(toolitem), browser->toolbar_location);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), toolitem, -1);

    // Search entry
    // TODO: custom icon(s); popup with search engines
    browser->toolbar_search = sexy_icon_entry_new();
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->toolbar_search), TRUE);
    /*sokoke_entry_set_default_text(GTK_ENTRY(browser->toolbar_search)
     , "Search the web");*/
    browser->icon_search = gtk_image_new_from_stock(GTK_STOCK_FIND
     , GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->toolbar_search)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(browser->icon_search));
    browser->icon_search_add = gtk_image_new();
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->toolbar_search)
     , SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE(browser->icon_search_add));
    g_signal_connect(browser->toolbar_search, "activate"
     , G_CALLBACK(on_toolbar_entry_search_activate), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->toolbar_search);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), toolitem, -1);

    browser->toolbar_closetab = tool_button_new(NULL
     , STOCK_TAB_CLOSE, FALSE, FALSE, G_CALLBACK(on_menu_file_close_tab_activate), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), browser->toolbar_closetab, -1);
    browser->toolbar_tabtrash = tool_button_new("Closed tabs"
     , STOCK_USER_TRASH, TRUE, FALSE, G_CALLBACK(on_toolbar_button_tabtrash_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->toolbar), browser->toolbar_tabtrash, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->toolbar));
    gtk_box_pack_start(GTK_BOX(vbox), browser->toolbar, FALSE, FALSE, 0);

    // Bookmarkbar
    browser->bookmarkbar = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->bookmarkbar), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->bookmarkbar), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = tool_button_new("Folder", GTK_STOCK_DIRECTORY
     , TRUE, TRUE, NULL, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK
     , TRUE, TRUE, NULL/*on_menu_bookmarks_item_activate*/, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    toolitem = tool_button_new("Example", STOCK_BOOKMARK
     , TRUE, TRUE, NULL/*on_menu_bookmarks_item_activate*/, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->bookmarkbar), toolitem, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->bookmarkbar));
    gtk_box_pack_start(GTK_BOX(vbox), browser->bookmarkbar, FALSE, FALSE, 0);

    // Superuser warning
    #ifdef G_OS_UNIX
    if(G_UNLIKELY(!geteuid())) // effective superuser?
    {
     hbox = gtk_event_box_new();
     gtk_widget_modify_bg(hbox, GTK_STATE_NORMAL
      , &hbox->style->bg[GTK_STATE_SELECTED]);
     label = gtk_label_new("Warning: You are using the superuser account!");
     gtk_misc_set_padding(GTK_MISC(label), 0, 2);
     gtk_widget_modify_fg(GTK_WIDGET(label), GTK_STATE_NORMAL
      , &GTK_WIDGET(label)->style->fg[GTK_STATE_SELECTED]);
     gtk_container_add(GTK_CONTAINER(hbox), GTK_WIDGET(label));
     gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    }
    #endif

    hbox = gtk_hpaned_new();
    // TODO: Load and save the hpaned's position
    // TODO: Does '0' actually mean 'minimum requested size'?
    gtk_paned_set_position(GTK_PANED(hbox), 0);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);
    browser->sidebars = gtk_vbox_new(FALSE, 0);
    gtk_paned_pack1(GTK_PANED(hbox), browser->sidebars, FALSE, FALSE);

    GtkWidget* cbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->sidebars), cbox, TRUE, TRUE, 0);

    // Sidebars title bar
    label = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(label), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(label), GTK_TOOLBAR_ICONS);
    toolitem = gtk_tool_item_new();
    gtk_tool_item_set_expand(toolitem, TRUE);
    // TODO: gtk_misc_set_alignment(GTK_MISC(widget), 0.0, 0.5);
    gtk_container_add(GTK_CONTAINER(toolitem)
     , gtk_label_new_with_mnemonic("_Sidebars"));
    gtk_toolbar_insert(GTK_TOOLBAR(label), toolitem, -1);
    // TODO: Does 'goto top' actually indicate 'make floating'?
    toolitem = tool_button_new(NULL, GTK_STOCK_GOTO_TOP, FALSE
     , TRUE, NULL/*G_CALLBACK(on_sidebars_button_float_clicked)*/, browser);
    gtk_toolbar_insert(GTK_TOOLBAR(label), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE
     , TRUE, G_CALLBACK(on_sidebars_button_close_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(label), toolitem, -1);
    gtk_box_pack_start(GTK_BOX(cbox), label, FALSE, FALSE, 0);

    // Notebook, containing all sidebars
    browser->sidebars_notebook = gtk_notebook_new();
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(browser->sidebars_notebook), TRUE);
      // Bookmarkbar
      GtkWidget* bookmarkbar = gtk_label_new("this is a tabbed sidebar");
      // TODO: Create bookmarkbar
      gtk_notebook_append_page(GTK_NOTEBOOK(browser->sidebars_notebook)
       , bookmarkbar, gtk_image_new_from_icon_name(STOCK_BOOKMARKS, GTK_ICON_SIZE_MENU));
      // History
      GtkWidget* historybar = gtk_label_new("webkit rocks");
      // TODO: Create historybar
      gtk_notebook_append_page(GTK_NOTEBOOK(browser->sidebars_notebook)
       , historybar, gtk_image_new_from_icon_name(STOCK_HISTORY, GTK_ICON_SIZE_MENU));
    // TODO: Add more sidebars here
    gtk_box_pack_start(GTK_BOX(cbox), browser->sidebars_notebook, TRUE, TRUE, 0);
    sokoke_container_show_children(GTK_CONTAINER(browser->sidebars));
    gtk_widget_show(browser->sidebars); //...

    // Notebook, containing all documents
    browser->pages = gtk_notebook_new();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(browser->pages), TRUE);
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_notebook_set_group_id(GTK_NOTEBOOK(browser->pages), 0);
    #endif
    g_signal_connect_after(GTK_OBJECT(browser->pages), "switch-page"
     , GTK_SIGNAL_FUNC(on_notebook_switch_page), browser);
    gtk_paned_pack2(GTK_PANED(hbox), browser->pages, FALSE, FALSE);
    gtk_widget_show(browser->pages);

    // Incremental findbar
    browser->findbox = gtk_toolbar_new();
    gtk_toolbar_set_icon_size(GTK_TOOLBAR(browser->findbox), GTK_ICON_SIZE_MENU);
    gtk_toolbar_set_style(GTK_TOOLBAR(browser->findbox), GTK_TOOLBAR_BOTH_HORIZ);
    toolitem = tool_button_new(NULL, GTK_STOCK_CLOSE, FALSE
     , TRUE, G_CALLBACK(on_findbox_button_close_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = gtk_tool_item_new();
    GtkStockItem stockitem; gtk_stock_lookup(GTK_STOCK_FIND, &stockitem);
    gtk_container_add(GTK_CONTAINER(toolitem)
     , gtk_label_new_with_mnemonic(stockitem.label));
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_text = sexy_icon_entry_new();
    GtkWidget* icon = gtk_image_new_from_stock(GTK_STOCK_FIND, GTK_ICON_SIZE_MENU);
    sexy_icon_entry_set_icon(SEXY_ICON_ENTRY(browser->findbox_text)
     , SEXY_ICON_ENTRY_PRIMARY, GTK_IMAGE(icon));
    sexy_icon_entry_add_clear_button(SEXY_ICON_ENTRY(browser->findbox_text));
    sokoke_entry_set_can_undo(GTK_ENTRY(browser->findbox_text), TRUE);
    g_signal_connect(browser->findbox_text, "activate"
     , GTK_SIGNAL_FUNC(on_findbox_button_down_clicked), browser);
    toolitem = gtk_tool_item_new();
    gtk_container_add(GTK_CONTAINER(toolitem), browser->findbox_text);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_DOWN, TRUE
     , TRUE, G_CALLBACK(on_findbox_button_down_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    toolitem = tool_button_new(NULL, GTK_STOCK_GO_UP, TRUE
     , TRUE, G_CALLBACK(on_findbox_button_up_clicked), browser);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), toolitem, -1);
    browser->findbox_case = gtk_toggle_tool_button_new_from_stock(GTK_STOCK_SPELL_CHECK);
    gtk_tool_button_set_label(GTK_TOOL_BUTTON(browser->findbox_case), "Match case");
    gtk_tool_item_set_is_important(GTK_TOOL_ITEM(browser->findbox_case), TRUE);
    gtk_toolbar_insert(GTK_TOOLBAR(browser->findbox), browser->findbox_case, -1);
    sokoke_container_show_children(GTK_CONTAINER(browser->findbox));
    gtk_box_pack_start(GTK_BOX(vbox), browser->findbox, FALSE, FALSE, 0);

    // TODO: fix children overlapping statusbar border
    browser->statusbar = gtk_statusbar_new();
    gtk_box_pack_start(GTK_BOX(vbox), browser->statusbar, FALSE, FALSE, 0);
    browser->progress = gtk_progress_bar_new();
    // Setting the progressbar's height to 1 makes it fit in the statusbar
    gtk_widget_set_size_request(browser->progress, -1, 1);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->progress
     , FALSE, FALSE, 3);

    browser->icon_link_script = gtk_image_new_from_icon_name(
     STOCK_SCRIPT, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_script
     , FALSE, FALSE, 0);
    browser->icon_link_window = gtk_image_new_from_icon_name(
     STOCK_WINDOW_NEW, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->icon_link_window
     , FALSE, FALSE, 0);

    //TODO: tooltips
    browser->misc_feed = gtk_hbox_new(FALSE, 0);
    icon = gtk_image_new_from_icon_name(GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(browser->misc_feed), icon, FALSE, FALSE, 0);
    label = gtk_label_new("News feed available");
    gtk_box_pack_start(GTK_BOX(browser->misc_feed), label, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(browser->statusbar), browser->misc_feed
     , FALSE, FALSE, 0);

    }
    else
    {

    browser->window = oldBrowser->window;
    browser->menubar = oldBrowser->menubar;
    browser->menu_closetab = oldBrowser->menu_closetab;
    browser->menu_zoom_in = oldBrowser->menu_zoom_in;
    browser->menu_zoom_out = oldBrowser->menu_zoom_out;
    browser->menu_zoom_normal = oldBrowser->menu_zoom_normal;
    browser->menu_go_back = oldBrowser->menu_go_back;
    browser->menu_go_forward = oldBrowser->menu_go_forward;
    browser->menu_window = oldBrowser->menu_window;
    browser->toolbar = oldBrowser->toolbar;
    browser->toolbar_back = oldBrowser->toolbar_back;
    browser->toolbar_forward = oldBrowser->toolbar_forward;
    browser->toolbar_refresh = oldBrowser->toolbar_refresh;
    browser->toolbar_closetab = oldBrowser->toolbar_closetab;
    browser->toolbar_tabtrash = oldBrowser->toolbar_tabtrash;
    browser->bookmarkbar = oldBrowser->bookmarkbar;
    browser->icon_page = oldBrowser->icon_page;
    browser->toolbar_location = oldBrowser->toolbar_location;
    browser->icon_lock = oldBrowser->icon_lock;
    browser->icon_search = oldBrowser->icon_search;
    browser->icon_search_add = oldBrowser->icon_search_add;
    browser->toolbar_search = oldBrowser->toolbar_search;
    browser->sidebars = oldBrowser->sidebars;
    browser->sidebars_notebook = oldBrowser->sidebars_notebook;
    browser->pages = oldBrowser->pages;
    browser->findbox = oldBrowser->findbox;
    browser->statusbar = oldBrowser->statusbar;
    browser->icon_link_script = oldBrowser->icon_link_script;
    browser->icon_link_window = oldBrowser->icon_link_window;
    browser->progress = oldBrowser->progress;
    browser->misc_feed = oldBrowser->misc_feed;

    }

    // Set the window features
    browser->bMenubar = TRUE;
    browser->bToolbar = TRUE;
    browser->bLocation = TRUE;
    browser->bStatusbar = TRUE;
    browser->bResizable = TRUE;

    // Add a window menu item
    // TODO: Menu items should be ordered like the notebook tabs
    browser->page_menu = menu_item_new("(untitled)", GTK_STOCK_FILE, "", NULL
     , (GdkModifierType)0, G_CALLBACK(on_menu_window_item_activate), TRUE, browser);
    gtk_menu_shell_append(GTK_MENU_SHELL(browser->menu_window), browser->page_menu);

    // Create a new page
    GtkWidget* eventbox = gtk_event_box_new();
    gtk_event_box_set_visible_window(GTK_EVENT_BOX(eventbox), FALSE);
    g_signal_connect(GTK_OBJECT(eventbox), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_mouse_down), browser);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 1);
    gtk_container_add(GTK_CONTAINER(eventbox), GTK_WIDGET(hbox));
    browser->page_icon = gtk_image_new_from_stock(GTK_STOCK_FILE, GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(hbox), browser->page_icon, FALSE, FALSE, 0);
    browser->page_name = gtk_label_new("(untitled)");
    gtk_misc_set_alignment(GTK_MISC(browser->page_name), 0.0, 0.5);
    // TODO: make the tab initially look "unvisited" until it's focused
    // TODO: tabs should shrink when there is not enough space
    // TODO: gtk's tab scrolling is weird?
    gint w, h; sokoke_widget_get_text_size(browser->page_name, "M", &w, &h);
    gtk_widget_set_size_request(GTK_WIDGET(browser->page_name)
     , w * config->iTabSize, -1);
    gtk_label_set_ellipsize(GTK_LABEL(browser->page_name), PANGO_ELLIPSIZE_END);
    gtk_box_pack_start(GTK_BOX(hbox), browser->page_name, FALSE, FALSE, 0);
    GtkWidget* button = gtk_button_new();
    gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);
    gtk_button_set_focus_on_click(GTK_BUTTON(button), FALSE);
    GtkRcStyle* rcstyle = gtk_rc_style_new();
    rcstyle->xthickness = rcstyle->ythickness = 0;
    gtk_widget_modify_style(button, rcstyle);
    gtk_rc_style_unref(rcstyle);
    GtkWidget* image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
    gtk_button_set_image(GTK_BUTTON(button), image);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    if(config->iTabClose) gtk_widget_show(button);
    gtk_widget_show_all(GTK_WIDGET(eventbox));
    g_signal_connect(GTK_OBJECT(button), "button_press_event"
     , GTK_SIGNAL_FUNC(on_notebook_tab_close_mouse_down), browser);
    #ifdef MIDORI_WEBKIT
    GtkWidget* scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GTK_WIDGET_SET_FLAGS(scrolled, GTK_CAN_FOCUS);
    #else
    GtkWidget* scrolled = gtk_event_box_new();
    #endif
    g_signal_connect(GTK_OBJECT(scrolled), "button_press_event"
     , GTK_SIGNAL_FUNC(on_document_mouse_down), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "popup-menu"
     , GTK_SIGNAL_FUNC(on_document_popup), browser);
    g_signal_connect(GTK_OBJECT(scrolled), "scroll-event"
     , GTK_SIGNAL_FUNC(on_document_scroll), browser);
    gint page = gtk_notebook_insert_page(GTK_NOTEBOOK(browser->pages)
     , scrolled, GTK_WIDGET(eventbox)
     , gtk_notebook_get_current_page(GTK_NOTEBOOK(browser->pages)) + 1);
    #if GTK_CHECK_VERSION(2, 10, 0)
    gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(browser->pages)
     , scrolled, TRUE);
    gtk_notebook_set_tab_detachable(GTK_NOTEBOOK(browser->pages)
     , scrolled, TRUE);
    #endif

    // Create a new document
    #define SIG_CONNECT(__sig, __func) g_signal_connect(\
     GTK_OBJECT(browser->document), __sig\
     , GTK_SIGNAL_FUNC(__func), browser);
    gchar* sAgentName = g_strconcat(PACKAGE_NAME, "/", PACKAGE_VERSION, NULL);
    #ifdef MIDORI_WEBI
    browser->document = webi_new();
    webi_set_emit_internal_status(WEBI(browser->document), TRUE); // crap
    gtk_widget_show_all(browser->document); // bug
    SIG_CONNECT("location", on_document_location_changed)
    SIG_CONNECT("title", on_document_title_changed)
    SIG_CONNECT("load-start", on_document_load_started)
    SIG_CONNECT("load-stop", on_document_load_finished)
    SIG_CONNECT("status", on_document_load_changed)
    SIG_CONNECT("set-device-type", on_document_device_changed)
    SIG_CONNECT("selection", on_document_selection_changed)
    SIG_CONNECT("req-js-prompt", on_document_request_script_prompt)
    SIG_CONNECT("req-auth-prompt", on_document_request_auth_prompt)
    SIG_CONNECT("req-new-window", on_document_window_open)
    SIG_CONNECT("show-window", on_document_window_show)
    SIG_CONNECT("set-window-properties", on_document_window_features)
    SIG_CONNECT("get-window-size", on_document_window_get_size)
    SIG_CONNECT("set-window-size", on_document_window_set_size)
    SIG_CONNECT("close-window", on_document_window_close)
    SIG_CONNECT("set-cookie", on_document_set_cookie)
    SIG_CONNECT("mouse-over", on_document_link_hover)
    #elif defined(MIDORI_WEBKIT)
    browser->document = GTK_WIDGET(webkit_gtk_page_new());
    /*webkit_gtk_page_set_application_name_for_user_agent(
     PACKAGE_NAME, PACKAGE_VERSION);*/
    //webkit_gtk_page_set_custom_user_agent(sAgentName);
    SIG_CONNECT("title-changed", on_document_title_changed)
    SIG_CONNECT("load-progress-changed", on_document_load_changed)
    SIG_CONNECT("status_bar_text_changed", on_document_status_message)
    #else
    webkit_view_set_application_name_for_user_agent(
     WEBKIT_VIEW(browser->document), sAgentName);
    browser->document = webkit_view_new(NULL, NULL);
    #endif
    g_free(sAgentName);

    SIG_CONNECT("destroy", on_document_destroy)
    #undef SIG_CONNECT

    // Eventually pack and display everything
    gtk_container_add(GTK_CONTAINER(scrolled), browser->document);
    if(TRUE/*config->bShowMenubar*/) gtk_widget_show_all(browser->menubar);
    if(config->bShowToolbar) gtk_widget_show(browser->toolbar);
    if(config->iTabClose) gtk_widget_hide(GTK_WIDGET(browser->toolbar_closetab));
    if(config->bShowBookmarkbar) gtk_widget_show(browser->toolbar);
    if(config->bShowStatusbar) gtk_widget_show(browser->statusbar);
    gtk_widget_show(GTK_WIDGET(scrolled));
    gtk_widget_show(GTK_WIDGET(browser->document));
    gtk_notebook_set_current_page(GTK_NOTEBOOK(browser->pages), page);

    update_page_actions(browser);
    gtk_widget_show(browser->window);
    gtk_widget_grab_focus(GTK_WIDGET(browser->toolbar_location));

    return browser;
}

// -- configuration related

static CConfig* config_load(const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    CConfig* config = 0;
    config = g_new0(CConfig, 1);

    GKeyFile* keyFile = g_key_file_new();
    g_key_file_load_from_file(keyFile, sFilename
     , G_KEY_FILE_KEEP_COMMENTS, error);
    if(1){
     #define GET_INT(var, key, default) \
      var = sokoke_key_file_get_integer_default( \
      keyFile, "browser", key, default, NULL)
     #define GET_STR(var, key, default) \
      var = sokoke_key_file_get_string_default( \
      keyFile, "browser", key, default, NULL)
     // TODO: Quit using constants which confuse humans?
     GET_STR(config->sHomepage, "Homepage", "http://www.google.com");
     GET_STR(config->sLocationSearch, "LocationSearch", "http://www.google.com/search?q=%s");
     GET_INT(config->bShowToolbar, "ShowNavigationbar", TRUE);
     GET_INT(config->bShowBookmarkbar, "ShowBookmarkbar", FALSE);
     GET_INT(config->bShowStatusbar, "ShowStatusbar", TRUE);
     //GET_INT(config->bShowDownloadbar, "ShowDownloadbar", FALSE);
     GET_INT(config->iToolbarSize, "ToolbarSize", GTK_ICON_SIZE_LARGE_TOOLBAR);
     GET_INT(config->iTabSize, "TabSize", 10);
     GET_INT(config->iTabClose, "TabClose", 1);
     GET_INT(config->iPopupRule, "PopupRule", 2);
     #undef GET_INT
     #undef GET_STR
    }

    return config;
}

gboolean config_save(CConfig* config, const gchar* sFilename, GError* *error)
{
    DEBUG_FUNC
    GKeyFile* keyFile = g_key_file_new();

    g_key_file_set_string(keyFile, "browser", "Homepage", config->sHomepage);
    g_key_file_set_string(keyFile, "browser", "LocationSearch", config->sLocationSearch);
    g_key_file_set_integer(keyFile, "browser", "ShowNavigationbar", config->bShowToolbar);
    g_key_file_set_integer(keyFile, "browser", "ShowBookmarkbar", config->bShowBookmarkbar);
    //g_key_file_set_integer(keyFile, "browser", "ShowDownloadbar", config->bShowDownloadbar);
    g_key_file_set_integer(keyFile, "browser", "ShowStatusbar", config->bShowStatusbar);
    g_key_file_set_integer(keyFile, "browser", "ToolbarSize", config->iToolbarSize);
    g_key_file_set_integer(keyFile, "browser", "TabSize", config->iTabSize);
    g_key_file_set_integer(keyFile, "browser", "TabClose", config->iTabClose);
    g_key_file_set_integer(keyFile, "browser", "PopupRule", config->iPopupRule);

    gboolean bSaved = sokoke_key_file_save_to_file(keyFile, sFilename, error);
    g_key_file_free(keyFile);

    return bSaved;
}

// -- stock icons

void init_stock_icons(void)
{
    // Based on ephy_stock_icons_init from lib/ephy-stock-icons.c from epiphany.
    // Copyright (C) 2002 Jorn Baayen
    GtkIconFactory *factory;
    GtkIconSet *icon_set;
    GtkIconSource *icon_source;
    int i;
    static const gchar* icon_theme_items[] =
    {
       STOCK_LOCK_OPEN
     , STOCK_LOCK_SECURE
     , STOCK_LOCK_BROKEN
     , STOCK_SCRIPT
     , STOCK_THEME
     , STOCK_USER_TRASH
    };
    static const GtkStockItem items[] =
    {
       { STOCK_BOOKMARK,       "Bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARK_NEW,   "New bookmark", 0, 0, NULL }
     , { STOCK_BOOKMARKS,      "_Bookmarks", 0, 0, NULL }
     , { STOCK_DOWNLOADS,      "_Downloads", 0, 0, NULL }
     , { STOCK_ERROR_CONSOLE,  "_Error console", 0, 0, NULL }
     , { STOCK_EXTENSIONS,     "_Extensions", 0, 0, NULL }
     , { STOCK_FORM_FILL,      "_Form fill", 0, 0, NULL }
     , { STOCK_HISTORY,        "History", 0, 0, NULL }
     , { STOCK_HOMEPAGE,       "Homepage", 0, 0, NULL }
     , { STOCK_LOCATION,       "Location entry", 0, 0, NULL }
     , { STOCK_NEWSFEED,       "Newsfeed", 0, 0, NULL }
     , { STOCK_PLUGINS,        "_Plugins", 0, 0, NULL }
     , { STOCK_POPUPS_BLOCKED, "Blocked popups", 0, 0, NULL }
     , { STOCK_SOURCE_VIEW,    "View source", 0, 0, NULL }
     , { STOCK_TAB_CLOSE,      "Close tab", 0, 0, NULL }
     , { STOCK_TAB_NEW,        "New tab", 0, 0, NULL }
     , { STOCK_WINDOW_CLOSE,   "Close window", 0, 0, NULL }
     , { STOCK_WINDOW_NEW,     "New window", 0, 0, NULL }
    };
    // TODO: Use fallback icons with _ suffix
    factory = gtk_icon_factory_new ();
    for(i = 0; i < (int)G_N_ELEMENTS(items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name (icon_source, items[i].stock_id);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, items[i].stock_id, icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_stock_add_static(items, G_N_ELEMENTS(items));

    for(i = 0; i < (int)G_N_ELEMENTS(icon_theme_items); i++)
    {
     icon_source = gtk_icon_source_new();
     gtk_icon_source_set_icon_name(icon_source, icon_theme_items[i]);
     icon_set = gtk_icon_set_new();
     gtk_icon_set_add_source(icon_set, icon_source);
     gtk_icon_source_free(icon_source);
     gtk_icon_factory_add(factory, icon_theme_items[i], icon_set);
     gtk_icon_set_unref(icon_set);
    }
    gtk_icon_factory_add_default(factory);
    g_object_unref(factory);

    // TODO: multi-head?
    // TODO: Do we need that stuff?
    /*gtk_icon_theme_append_search_path(gtk_icon_theme_get_default()
     , SHARE_DIR G_DIR_SEPARATOR_S "icons");*/
}

// -- crash safe

#if HAVE_UNISTD_H && HAVE_STDLIB_H
RETSIGTYPE signal_crash_handler(int signal)
{
    #ifdef MIDORI_CRASHSAFE
    // TODO: backup session
    #endif
    g_print("Terminated: %s\n", g_strsignal(signal));
    g_print("%s\n", strlen(DEBUG_MSG) ? DEBUG_MSG : "<unknown>");
    exit(1);
}
#endif

// -- main function

int main(int argc, char* *argv)
{
    DEBUG_FUNC
    // TODO: Any way to have a crash handler on non-posix platforms?
    #if HAVE_UNISTD_H && HAVE_STDLIB_H
    signal(SIGSEGV, signal_crash_handler);
    signal(SIGABRT, signal_crash_handler);
    #endif

    // Parse cli options
    gint repeats = 2;
    gboolean bVersion = FALSE;
    GOptionEntry entries[] =
    {
     { "repeats", 'r', 0, G_OPTION_ARG_INT, &repeats, "An unused value", "N" },
     { "version", 'v', 0, G_OPTION_ARG_NONE, &bVersion, "Display program version", NULL }
    };
    GError* error = NULL;
    if(!gtk_init_with_args(&argc, &argv, "[URI]", entries
     , NULL/*GETTEXT_PACKAGE*/, &error)) return 1; g_free(error);

    if(bVersion)
    {
     g_print(PACKAGE_STRING " - Copyright (c) 2007 Christian Dywan\n\n"
             "GTK+2:      " GTK_VER "\n"
	     "WebKit:     " WEBKIT_VER "\n"
	     "Libsexy:    " LIBSEXY_VER "\n"
             "GetText:    N/A\n"
	     "\n"
	     "CrashSafe:  " MIDORI_CRASHSAFE_ "\n"
	     "Debugging:  " MIDORI_DEBUG_ "\n"
	     "Engine:     " MIDORI_ENGINE_ "\n"
             "\n"
             "Please report comments, suggestions and bugs to:\n"
             "\t" PACKAGE_BUGREPORT "\n"
             "Check for new versions at:\n"
             "\thttp://software.twotoasts.de\n");
     return 0;
    }

    // Load configuration files
    // TODO: What about default config in a global config folder?
    gchar* sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    gchar* sConfigFile = g_build_filename(sConfig, "config", NULL);
    config = config_load(sConfigFile, &error); g_free(sConfigFile);
    if(error != NULL)
    {
     g_print("Notice: No configuration loaded. %s", error->message);
     g_free(error);
    }
    gchar* sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_load(sAccelsFile); g_free(sAccelsFile);
    /*gchar* sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    bookmarks = g_bookmark_file_new();
    if(!g_bookmark_file_load_from_file(bookmarks, sBookmarkFile, &error))
     g_warning("Bookmarks couldn't be loaded. %s", error->message); g_free(sBookmarkFile);*/
    g_free(sConfig);

    browser_list = g_list_alloc();
    tabtrash = NULL;
    accel_group = gtk_accel_group_new();

    // TODO: Use custom program icon
    gtk_window_set_default_icon_name("web-browser");
    init_stock_icons();
    #ifdef MIDORI_WEBKIT
    webkit_gtk_init();
    #endif

    gchar* sHomepage = argc > 1 ? argv[1] : config->sHomepage;
    // Open as many tabs as we have uris, seperated by pipes
    // TODO: This is not utf8 save
    CBrowser* browser = NULL;
    gchar* sUri = strtok(g_strdup(sHomepage), "|");
    while(sUri != NULL)
    {
     browser = browser_new(browser);
     // TODO: Is it really wise to call this from outside the main loop?
     document_load_uri(browser->document, magic_uri(sUri));
     sUri = strtok(NULL, "|");
    } g_free(sUri);

    gtk_main();
    DEBUG_FUNC

    // Save configuration files
    sConfig = g_build_filename(g_get_user_config_dir(), PACKAGE_NAME, NULL);
    g_mkdir_with_parents(sConfig, 0755);
    sConfigFile = g_build_filename(sConfig, "config", NULL);
    if(!config_save(config, sConfigFile, &error))
     g_warning("Configuration couldn't be saved. %s", error->message); g_free(sConfigFile);
    sAccelsFile = g_build_filename(sConfig, "accels", NULL);
    gtk_accel_map_save(sAccelsFile); g_free(sAccelsFile);
    /*sBookmarkFile = g_build_filename(sConfig, "bookmarks.xbel", NULL);
    if(!g_bookmark_file_to_file(bookmarks, sBookmarkFile, NULL))
     g_warning("Bookmarks couldn't be saved."); g_free(sBookmarkFile);*/
    g_free(sConfig);
    return 0;
}
