/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#ifndef __MIDORI_H__
#define __MIDORI_H__ 1

#include "config.h"

#ifdef HAVE_STDLIB_H
 #include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #include <signal.h>
#endif
// TODO: Including specific gtk headers should be better
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libsexy/sexy-icon-entry.h>

#include "sokoke.h"

#ifdef MIDORI_WEBI
 #include <webi/webi.h>
#elif defined(MIDORI_WEBKIT)
 #include <webkitgtkpage.h>
 #include <webkitgtkglobal.h>
#else
 #include <webkit-gtk/webkit.h>
 #include <webi-external-content-delegate.h>
 #include <webi-auth.h>
 #include <webkit-gtk/webkit-view-webi-extensions.h>
#endif

// Our beautiful debug macros go here

static gchar DEBUG_MSG[256];
#ifndef __func__
 #define __func__ __FUNCTION__
#endif
#if MIDORI_DEBUG > 1
 #define DEBUG_FUNC g_snprintf(DEBUG_MSG, 255, "%s %s %d"\
  , __FILE__, __func__, __LINE__);g_print("%s\n", DEBUG_MSG);
#elif MIDORI_DEBUG > 0
 #define DEBUG_FUNC g_snprintf(DEBUG_MSG, 255, "%s %s %d"\
  , __FILE__, __func__, __LINE__);
#else
 #define DEBUG_FUNC ;
#endif

// For unimplemeted functions
#define UNIMPLEMENTED g_print("Unimplemented: %s\n", __func__);

// Many themes need this hack for small toolbars to work
#define GTK_ICON_SIZE_SMALL_TOOLBAR GTK_ICON_SIZE_MENU

// Miscellaneous constants

#define MIDORI_LEN_URI 255

// Custom stock items

// We should distribute these
// Names should match with epiphany and/ or xdg spec

#define STOCK_BOOKMARK        "stock_bookmark" // "bookmark-web"
#define STOCK_BOOKMARKS       "bookmark-view"
#define STOCK_DOWNLOADS       "package" // "download"
#define STOCK_ERROR_CONSOLE   "error-console" // MISSING
#define STOCK_EXTENSIONS      "extension" // MISSING
#define STOCK_FORM_FILL       "form-fill" // MISSING
#define STOCK_HISTORY         "history-view"
#define STOCK_LOCATION        "location-entry"
#define STOCK_NEWSFEED        "newsfeed" // MISSING
#define STOCK_PLUGINS         "plugins" // MISSING
#define STOCK_POPUPS_BLOCKED  "popup-hidden"
#define STOCK_TAB_CLOSE       "tab-close" // MISSING
#define STOCK_WINDOW_CLOSE    "window-close" // MISSING

// We can safely assume to have those

#define STOCK_BOOKMARK_NEW    "bookmark-new"
#define STOCK_BOOKMARK_NEW_   "stock_add-bookmark"
#define STOCK_HOMEPAGE        GTK_STOCK_HOME
#define STOCK_IMAGE           "image-x-generic"
#define STOCK_IMAGE_          "gnome-mime-image"
#define STOCK_LOCK_OPEN       "stock_lock-open"
#define STOCK_LOCK_SECURE     "stock_lock"
#define STOCK_LOCK_BROKEN     "stock_lock-broken"
#define STOCK_SCRIPT          "stock_script"
#define STOCK_TAB_NEW         "tab-new"
#define STOCK_TAB_NEW_        "stock_new-tab"
#define STOCK_THEME           "gnome-settings-theme"
#define STOCK_SOURCE_VIEW     "stock_view-html-source"
#define STOCK_USER_TRASH      "user-trash"
#define STOCK_USER_TRASH_     "gnome-stock-trash"
#define STOCK_WINDOW_NEW      "window-new"
#define STOCK_WINDOW_NEW_     "stock_new-window"

// For backwards compatibility

#ifndef GTK_STOCK_SELECT_ALL
#define GTK_STOCK_SELECT_ALL  "gtk-select-all"
#endif

// -- Browser object

typedef struct _CBrowser
{
    // shared widgets
    GtkWidget* window;
    GtkWidget* menubar;
    GtkWidget* menu_closetab;
    GtkWidget* menu_zoom_in;
    GtkWidget* menu_zoom_out;
    GtkWidget* menu_zoom_normal;
    GtkWidget* menu_go_back;
    GtkWidget* menu_go_forward;
    GtkWidget* menu_window;
    //GtkWidget* menu_tabs;
    GtkWidget* toolbar;
    GtkToolItem* toolbar_back;
    GtkToolItem* toolbar_forward;
    GtkToolItem* toolbar_refresh;
    GtkToolItem* toolbar_closetab;
    GtkToolItem* toolbar_tabtrash;
    GtkWidget* bookmarkbar;
    GtkWidget* icon_page;
    GtkWidget* toolbar_location;
    GtkWidget* icon_lock;
    GtkWidget* icon_search;
    GtkWidget* icon_search_add;
    GtkWidget* toolbar_search;
    GtkWidget* sidebars;
    GtkWidget* sidebars_notebook;
    GtkWidget* pages;
    GtkWidget* findbox;
    GtkWidget* findbox_text;
    GtkToolItem* findbox_case;
    GtkWidget* statusbar;
    GtkWidget* icon_link_script;
    GtkWidget* icon_link_window;
    GtkWidget* progress;
    GtkWidget* misc_feed;

    // tab specific widgets
    GtkWidget* page_menu;
    GtkWidget* page_icon;
    GtkWidget* page_name;
    GtkWidget* document;

    // tab specific values
    gboolean bMenubar;
    gboolean bToolbar;
    gboolean bLocation;
    gboolean bStatusbar;
    gboolean bResizable;
    gint loadPercent;
    gint bytesLoaded;
    gint maxBytesLoaded;
    //UNDEFINED favicon;
    guint security;
    gchar* status_message; // message from a document
    gchar** resources; // list of loaded uris
} CBrowser;

// -- Resource object

typedef struct _CResource
{
    gchar* sUri;
    gchar* sTitle;
} CResource;

// -- User configuration object

typedef struct _CConfig
{
    gchar* sHomepage;
    gchar* sLocationSearch;
    //gboolean bShowMenubar;
    gboolean bShowToolbar;
    gboolean bShowBookmarkbar;
    //gboolean bShowDownloadbar;
    gboolean bShowStatusbar;
    gint iToolbarSize;
    guint iTabSize; // tab size in charcters
    guint iTabClose; // 0: none, 1: every tab
    guint iPopupRule; // -> on_document_window_open
} CConfig;

// -- credits

 // TODO: Retrieve credits from some kind of normal text files
 const gchar* credits_authors[] = { "Christian Dywan <christian@twotoasts.de>", NULL };
 const gchar* credits_documenters/*[]*/ = /*{ */NULL/* }*/;
 const gchar* credits_artists[] = { "Nancy Runge <nancy@twotoasts.de>", NULL };

// -- licenses

const gchar* license_gpl_2 = \
 ("This program is free software; you can redistribute it and/or modify "
  "it under the terms of the GNU General Public License as published by "
  "the Free Software Foundation; either version 2 of the License, or "
  "(at your option) any later version."
  "\n\n"
  "This program is distributed in the hope that it will be useful, "
  "but WITHOUT ANY WARRANTY; without even the implied warranty of "
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
  "GNU General Public License for more details."
  "\n\n"
  "You should have received a copy of the GNU General Public License "
  "along with this program; if not, write to the Free Software Foundation, Inc., "
  "51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.");

// -- function declarations begin here

gboolean document_can_go_back(GtkWidget*);
gboolean document_can_go_forward(GtkWidget*);
G_CONST_RETURN gchar* document_get_uri(GtkWidget*);
G_CONST_RETURN gchar* document_get_title(GtkWidget*);
gfloat document_get_text_size(GtkWidget* document);
void on_document_popup(GtkWidget*, GdkEventButton*, CBrowser*);
static CBrowser* browser_new(CBrowser*);

#endif /* !__MIDORI_H__ */
